# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module provides data structures to represent the meta-data of a CZModel containing a TensorFlow ANN model."""
import json
import os
import uuid
from typing import Dict, List, Tuple, NamedTuple, Union, Optional, TYPE_CHECKING
from xml.etree.ElementTree import Element, SubElement, ElementTree

from .util.color_generation import ColorGenerator

# Ignore flake8 imported but unused warnings (does not recognize non-runtime check). Checked by mypy
if TYPE_CHECKING:
    from tensorflow.keras import Model  # noqa: F401


class SegmentationClass(NamedTuple):
    """Class representing the metadata of one target class of a segmentation problem."""

    name: str
    color: Tuple[int, int, int]

    def to_xml(self, label_value: int, node_name: str = "Item") -> Element:
        """Create an XML representation of the model meta data.

        Args:
            label_value: The value of the label assigned to this class.
            node_name: The name of the created node where the class metadata resides in.

        Returns:
            An XML node containing the class metadata.
        """
        # Define node
        model_node = Element(node_name)

        # Add attributes
        model_node.set("colB", str(self.color[2]))
        model_node.set("colG", str(self.color[1]))
        model_node.set("colR", str(self.color[0]))
        model_node.set("LabelValue", str(label_value))
        model_node.set("Name", self.name)

        return model_node


class ModelMetadata(NamedTuple):
    """Class representing the metadata of a model.

    Attributes:
        name: The name of the generated model.
        border_size: For Intellesis models this attribute defines the size of the border that needs to be added to an
            input image such that there are no border effects visible in the required area of the generated
            segmentation mask. For deep architectures this value can be infeasibly large so that the border size must
            be defined in a way that the border effects are “acceptable” in the ANN model creator’s opinion.
        color_handling (Type: string)**: Specifies how color (RGB and RGBA) pixel data are converted to one or more
            channels of scalar pixel data.
            Possible values are: ConvertToMonochrome (Converts color to gray scale),
            SplitRgb (Keeps the pixel representation in RGB space).
        pixel_type: The expected input type of the model. Possible values are:
            Gray8: 8 bit unsigned
            Gray16: 16 bit unsigned
            Gray32Float: 4 byte IEEE float
            Bgr24: 8 bit triples, representing the color channels Blue, Green and Red
            Bgr48: 16 bit triples, representing the color channels Blue, Green and Red
            Bgr96Float: Triple of 4 byte IEEE float, representing the color channels Blue, Green and Red
            Bgra32: 8 bit triples followed by an alpha (transparency) channel
            Gray64ComplexFloat: 2 x 4 byte IEEE float, representing real and imaginary part of a complex number
            Bgr192ComplexFloat: A triple of 2 x 4 byte IEEE float, representing real and imaginary part of a
                complex number, for the color channels Blue, Green and Red
        classes: A list of class names corresponding to the output dimensions of the predicted segmentation mask.
            If the last dimension of the prediction has shape n the provided list must be of length n.
        test_image_file: The path to a test image in a format supported by ZEN. This image is used for basic validation
            of the converted model inside ZEN. Can be absolute or relative to the JSON file.
        license_file: The path to a license file that is added to the generated CZModel. Can be absolute or relative to
            the JSON file.
    """

    name: str
    color_handling: str
    pixel_type: str
    classes: List[SegmentationClass]
    border_size: int
    license_file: Optional[str] = None
    post_processing: Optional[str] = None
    model_id: uuid.UUID = uuid.uuid4()

    @classmethod
    def from_json(cls, model_metadata: Union[str, Dict]) -> "ModelMetadata":
        """This function parses a model specification JSON file to a ModelMetadata object.

        Args:
            model_metadata: The path to the JSON file containing the model specification.

        Returns:
            A ModelMetadata instance carrying all information to generate a CZModel file.
        """
        spec: Dict
        # Load spec from json if necessary
        if isinstance(model_metadata, str):
            with open(model_metadata, "r") as f:
                spec = json.load(f)
        else:
            spec = model_metadata

        # Create color generator
        color_gen = iter(ColorGenerator())

        # Resolve classes
        classes_resolved = [
            SegmentationClass(seg_class, color)
            for seg_class, color in zip(spec["Classes"], color_gen)
        ]

        return ModelMetadata(
            name=os.path.basename(spec.get("ModelPath", "DNN model")),
            color_handling=spec["ColorHandling"],
            pixel_type=spec["PixelType"],
            classes=classes_resolved,
            border_size=spec["BorderSize"],
            license_file=spec.get("LicenseFile", None),
        )

    @classmethod
    def from_params(
        cls,
        color_handling: str,
        pixel_type: str,
        classes: List[str],
        border_size: int,
        name: str = "DNN Model",
        license_file: Optional[str] = None,
    ) -> "ModelMetadata":
        """This function creates ModelMetadata object for a given TensorFlow Keras model.

        Args:
            name: The name of the generated model.
            border_size: For Intellesis models this attribute defines the size of the border that needs to be added to
                an input image such that there are no border effects visible in the required area of the generated
                segmentation mask. For deep architectures this value can be infeasibly large so that the border size
                must be defined in a way that the border effects are “acceptable” in the ANN model creator’s opinion.
            color_handling: Specifies how color (RGB and RGBA) pixel data are converted to one or more channels of
                scalar pixel data. Possible values are: ConvertToMonochrome (Converts color to gray scale),
                SplitRgb (Keeps the pixel representation in RGB space).
            pixel_type: The expected input type of the model. Possible values are:
                Gray8: 8 bit unsigned
                Gray16: 16 bit unsigned
                Gray32Float: 4 byte IEEE float
                Bgr24: 8 bit triples, representing the color channels Blue, Green and Red
                Bgr48: 16 bit triples, representing the color channels Blue, Green and Red
                Bgr96Float: Triple of 4 byte IEEE float, representing the color channels Blue, Green and Red
                Bgra32: 8 bit triples followed by an alpha (transparency) channel
                Gray64ComplexFloat: 2 x 4 byte IEEE float, representing real and imaginary part of a complex number
                Bgr192ComplexFloat: A triple of 2 x 4 byte IEEE float, representing real and imaginary part of a
                    complex number, for the color channels Blue, Green and Red
            classes: A list of class names corresponding to the output dimensions of the predicted segmentation mask.
                If the last dimension of the prediction has shape n the provided list must be of length n.
            license_file: The path to a license file that is added to the generated CZModel. Can be absolute or
                relative to the JSON file.

        Returns:
            A ModelMetadata instance carrying all information to generate a CZModel file.
        """
        # Create color generator
        color_gen = iter(ColorGenerator())
        # Resolve classes
        classes_resolved = [
            SegmentationClass(seg_class, color)
            for seg_class, color in zip(classes, color_gen)
        ]

        model_metadata = ModelMetadata(
            name=name,
            color_handling=color_handling,
            pixel_type=pixel_type,
            classes=classes_resolved,
            border_size=border_size,
            license_file=license_file,
        )

        return model_metadata

    def to_xml(self) -> ElementTree:
        """Creates an XML representation of the model meta data."""
        # Create root node
        model_node = Element("Model")
        model_node.set("Version", "3.1.0")

        # Create model ID
        id_node = SubElement(model_node, "Id")
        id_node.text = str(self.model_id)

        # Add model name
        name_node = SubElement(model_node, "ModelName")
        name_node.text = str(self.name)

        # Add status (only 'Trained' is supported for .pb models)
        status_node = SubElement(model_node, "Status")
        status_node.text = "Trained"

        # Add feature extractor (only 'DeepNeuralNetwork' is supported)
        feature_extractor_node = SubElement(model_node, "FeatureExtractor")
        feature_extractor_node.text = "DeepNeuralNetwork"

        # Add post-processing node
        post_processing_node = SubElement(model_node, "Postprocessing")
        post_processing_node.text = self.post_processing

        # Add color handling node
        color_handling_node = SubElement(model_node, "ColorHandling")
        color_handling_node.text = self.color_handling

        # Add channels (currently only a single channel is supported
        channels_node = SubElement(model_node, "Channels")
        channels_item_node = SubElement(channels_node, "Item")
        channels_item_node.set("PixelType", self.pixel_type)

        # Add Segmentation Classes
        training_classes_node = SubElement(model_node, "TrainingClasses")
        for i, segmentation_class in enumerate(self.classes):
            training_classes_node.append(segmentation_class.to_xml(label_value=i + 1))

        # Add border size
        border_size_node = SubElement(model_node, "BorderSize")
        border_size_node.text = str(self.border_size)

        return ElementTree(model_node)


class ModelSpec(NamedTuple):
    """Model specification tuple."""

    model: Union[str, "Model"]
    model_metadata: ModelMetadata

    @classmethod
    def from_json(cls, model_spec_path: str) -> "ModelSpec":
        """This function parses a model specification JSON file to a ModelSpec object.

        Args:
            model_spec_path: The path to the JSON file containing the model specification.

        Returns:
            A ModelMetadata instance carrying all information to generate a CZModel file.
        """
        with open(model_spec_path, "r") as f:
            spec = json.load(f)

        return ModelSpec(
            model=spec["ModelPath"], model_metadata=ModelMetadata.from_json(spec)
        )



# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides preprocessing utilities."""
from typing import Any  # noqa # pylint: disable=unused-import
from typing import Optional, Sequence, Tuple, Union
import collections.abc
import warnings

import tensorflow as tf


class RgbToBgr(tf.keras.layers.Layer):
    """A Keras pre-processing layer that converts RGB inputs to BGR color representation."""

    @tf.function
    def call(self, inputs: "tf.Tensor", **kwargs: "Any") -> "tf.Tensor":
        """Performs the RGB to BGR conversion.

        Args:
            inputs: The RGB image.
            kwargs: Additional keyword arguments.

        Returns:
            The BGR image.
        """
        # pylint: disable=no-self-use
        return inputs[..., ::-1]


class RgbaToBgra(tf.keras.layers.Layer):
    """A Keras pre-processing layer that converts RGBA inputs to BGRA color representation."""

    @tf.function
    def call(self, inputs: "tf.Tensor", **kwargs: "Any") -> "tf.Tensor":
        """Performs the RGBA to BGRA conversion.

        Args:
            inputs: The RGBA image.
            kwargs: Additional keyword arguments.

        Returns:
            The BGRA image.
        """
        # pylint: disable=no-self-use
        return tf.concat([inputs[..., :-1][..., ::-1], inputs[..., -1:]], -1)


class PerImageStandardization(tf.keras.layers.Layer):
    """A Keras pre-processing layer that applies per image standardization."""

    @tf.function
    def call(self, inputs: "tf.Tensor", **kwargs: "Any") -> "tf.Tensor":
        """Shifts and linearly scales each image.

         The image will have mean 0 and variance 1. The image is implicitly converted to float representation.

        Args:
            inputs: The image(s) to be standardized.
            kwargs: Additional keyword arguments.

        Returns:
            The standardized image.
        """
        # pylint: disable=no-self-use
        if inputs.dtype != tf.float32:
            inputs = tf.image.convert_image_dtype(inputs, dtype=tf.float32)

        return tf.image.per_image_standardization(inputs)


class Normalization(tf.keras.layers.Layer):
    """A Keras pre-processing layer that shifts the data to zero mean and standard deviation 1."""

    def __init__(
        self,
        mean: Tuple[float, ...],
        std: Tuple[float, ...],
        *args: "Any",
        **kwargs: "Any"
    ) -> None:
        """Creates an instance of the normalization layer.

        Args:
            mean: The mean of the dataset.
            std: The standard deviation of the dataset.
            args: Additional arguments to be passed to the base Layer class.
            kwargs: Additional keyword arguments to be passed to the base Layer class.
        """
        super().__init__(*args, **kwargs)
        self._mean = tf.convert_to_tensor(mean)
        self._std = tf.convert_to_tensor(std)
        if tf.reduce_any(self._std == 0):
            warnings.warn(
                "The defined standard deviation contains zeros. This may lead to numerical issues."
            )

    @tf.function
    def call(self, inputs: "tf.Tensor", **kwargs: "Any") -> "tf.Tensor":
        """Shifts and linearly scales an image by constant values.

        Args:
            inputs: The image(s) to be normalized.
            kwargs: Additional keyword arguments.

        Returns:
            The normalized image.
        """
        return (inputs - tf.cast(self._mean, inputs.dtype)) / tf.cast(
            self._std, inputs.dtype
        )


DEFAULT_LAYER = RgbToBgr()


def add_preprocessing_layers(
    model: "tf.keras.Model",
    layers: Optional[
        Union["tf.keras.layers.Layer", Sequence["tf.keras.layers.Layer"]]
    ] = DEFAULT_LAYER,
    spatial_dims: Optional[Tuple["int", "int"]] = None,
) -> "tf.keras.Model":
    """Prepends a given pre-processing layer to a given Keras model.

    Args:
        model: The Keras model to be wrapped.
        layers: The layers to be prepended.
        spatial_dims: Set new spatial dimensions for the input node. This parameter is expected to contain the
            new height and width in that order. Note: Setting this parameter is only possible for models
            that are invariant to the spatial dimensions of the input such as FCNs.

    Returns:
        A new Keras model wrapping the provided Keras model and the pre-processing layers.
    """
    # Handle single layer and None input
    if layers is None:
        layers = []
    elif not isinstance(layers, collections.abc.Sequence):
        layers = [layers]

    # Create input layer
    new_shape = (
        (model.inputs[0].shape[0],) + spatial_dims + (model.inputs[0].shape[-1],)
        if spatial_dims is not None
        else model.inputs[0].shape
    )
    input_layer = tf.keras.Input(batch_shape=new_shape, name="input")

    # Apply pre-processing layer
    converted = input_layer
    for layer in layers:
        converted = layer(converted)

    # Apply model
    outputs = model(converted)

    # Return new Keras model
    return tf.keras.Model(inputs=input_layer, outputs=outputs)



# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides conversion functions to generate a CZModel from exported TensorFlow models."""
import os
from typing import Tuple, Union, Sequence, Optional, TYPE_CHECKING

from tensorflow.keras import Model
from tensorflow.keras.models import load_model

from .model_metadata import ModelSpec
from .util import keras_to_czmodel
from .util.argument_parsing import dir_file

# Ignore flake8 imported but unused warnings (does not recognize non-runtime check). Checked by mypy
if TYPE_CHECKING:
    from tensorflow.keras.layers import Layer  # noqa: F401


def convert_from_model_spec(
    model_spec: ModelSpec,
    output_path: str,
    output_name: str = "DNNModel",
    spatial_dims: Optional[Tuple[int, int]] = None,
    preprocessing: Union["Layer", Sequence["Layer"]] = None,
) -> None:
    """Convert a TensorFlow .pb or TensorFlow Keras model to a CZModel usable in ZEN Intellesis.

    Args:
        model_spec: A ModelSpec object describing the specification of the CZModel to be generated.
        output_path: A folder to store the generated CZModel file.
        output_name: The name of the generated .czmodel file.
        spatial_dims: New spatial dimensions for the input node (see final usage for more details)
        preprocessing: A sequence of layers to be prepended to the model. (see final usage for more details)
    """
    # Create output directory
    os.makedirs(output_path, exist_ok=True)

    # Load model if necessary
    model = (
        model_spec.model
        if isinstance(model_spec.model, Model)
        else load_model(model_spec.model)
    )

    # Convert model
    keras_to_czmodel.convert(
        model=model,
        model_metadata=model_spec.model_metadata,
        output_path=os.path.join(output_path, output_name),
        spatial_dims=spatial_dims,
        preprocessing=preprocessing,
    )


def convert_from_json_spec(
    model_spec_path: str,
    output_path: str,
    output_name: str = "DNNModel",
    spatial_dims: Optional[Tuple[int, int]] = None,
    preprocessing: Union["Layer", Sequence["Layer"]] = None,
) -> None:
    """Converts a TensorFlow SavedModel specified in a JSON metadata file to a CZModel that is usable in ZEN Intellesis.

    Args:
        model_spec_path: The path to the JSON specification file.
        output_path: A folder to store the generated CZModel file.
        output_name: The name of the generated .czmodel file.
        spatial_dims: New spatial dimensions for the input node (see final usage for more details)
        preprocessing: A sequence of layers to be prepended to the model. (see final usage for more details)
    """
    # Parse the specification JSON file
    parsed_spec = ModelSpec.from_json(model_spec_path)

    # Write CZModel to disk
    convert_from_model_spec(
        parsed_spec,
        output_path,
        output_name,
        spatial_dims=spatial_dims,
        preprocessing=preprocessing,
    )


def main() -> None:
    """Console script to convert a TensorFlow proto-buffer to a CZModel."""
    # Import argument parser
    import argparse  # pylint: disable=import-outside-toplevel

    # Define expected arguments
    parser = argparse.ArgumentParser(
        description="Convert a TensorFlow proto-buffer to a CZModel that can be executed inside ZEN."
    )
    parser.add_argument(
        "model_spec",
        type=dir_file,
        help="A JSON file containing the model specification.",
    )
    parser.add_argument(
        "output_path",
        type=str,
        help="The path where the generated CZModel will be created.",
    )
    parser.add_argument(
        "output_name", type=str, help="The name of the generated CZModel."
    )

    # Parse arguments
    args = parser.parse_args()

    # Run conversion
    convert_from_json_spec(args.model_spec, args.output_path, args.output_name)


if __name__ == "__main__":
    main()



# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides a color generator class that creates an infinite sequence of colors."""
import colorsys
from enum import Enum
from typing import Tuple, List, Optional


class Hue(Enum):
    """Container for basic colors."""

    RED = 0
    YELLOW = 1
    GREEN = 2
    CYAN = 3
    BLUE = 4
    MAGENTA = 5


HUE_LUT = {
    Hue.RED: Hue.GREEN,
    Hue.YELLOW: Hue.MAGENTA,
    Hue.GREEN: Hue.BLUE,
    Hue.CYAN: Hue.RED,
    Hue.BLUE: Hue.YELLOW,
    Hue.MAGENTA: Hue.CYAN,
}


class ColorGenerator:
    """Class providing functionality for generating an infinite sequence of colors."""

    def __init__(self, initial_color: Tuple[int, int, int] = (255, 0, 0)) -> None:
        """Initializes a color generator instance.

        Args:
            initial_color: The color to start with.
        """
        self.initial_color = colorsys.rgb_to_hls(
            *(initial_color_channel / 255.0 for initial_color_channel in initial_color)
        )
        self.current_color: Optional[List[float]] = None
        self.current_iteration = 0

    def __iter__(self) -> "ColorGenerator":
        """Creates an iterator from the color generator class."""
        return self

    def __next__(self) -> Tuple[int, int, int]:
        """Yields the next color."""
        if self.current_color is None:
            self.current_color = list(self.initial_color)
        else:
            # Make sure that the hue influences the color perceivably
            if self.current_color[2] < 0.5:
                self.current_color[2] = 0.5

            if self.current_color[1] < 0.3:
                self.current_color[1] = 0.3
            elif self.current_color[1] > 0.7:
                self.current_color[1] = 0.7

            # To prevent pink directly after red
            if self.current_iteration == 1:
                self.current_iteration += 1

            next_hue = float(HUE_LUT[Hue(self.current_iteration % len(Hue))].value)

            iteration = int(self.current_iteration / len(Hue))
            next_hue += iteration / (iteration + 1.0)
            self.current_color[0] = ((next_hue * (1.0 / len(Hue))) + 1.0) % 1.0

            self.current_iteration += 1

        color_rgb = colorsys.hls_to_rgb(*self.current_color)
        return int(color_rgb[0] * 255), int(color_rgb[1] * 255), int(color_rgb[2] * 255)



# Copyright 2021 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides common functionality for packing ANN models into a czmodel file."""
import os
from io import BytesIO
from typing import Union, TYPE_CHECKING
from zipfile import ZipFile

# Ignore flake8 imported but unused warnings (does not recognize non-runtime check). Checked by mypy
if TYPE_CHECKING:
    from czmodel.model_metadata import ModelMetadata  # noqa: F401


def zip_directory(directory: str, zip_file: ZipFile) -> None:
    """Adds an entire directory to a ZipFile.

    Args:
        directory: The directory to be added to the zip file.
        zip_file: The zip file to add the directory to.
    """
    # Iterate all the files in directory
    for folder, _, filenames in os.walk(directory):
        for filename in filenames:
            # Determine full path
            full_path = os.path.join(folder, filename)
            # Determine path relative to directory
            rel_path = os.path.relpath(full_path, directory)
            # Add file to zip
            zip_file.write(full_path, arcname=rel_path)


def create_model_zip(
    model: Union[str, bytes, memoryview],
    model_metadata: "ModelMetadata",
    output_path: str,
) -> None:
    """Creates a CZModel file from a given ModelMetadata.

    Args:
        model: The Keras model to be packed.
        model_metadata: The meta data describing the CZModel to be generated.
        output_path: The path of the CZModel file to be generated.
    """
    # Append correct extension if necessary
    if not output_path.lower().endswith(".czmodel"):
        output_path = output_path + ".czmodel"

    # Create ZIP file
    with ZipFile(output_path, mode="w") as zf:
        # Write model xml file
        buffer = BytesIO()
        model_metadata.to_xml().write(buffer, encoding="utf-8", xml_declaration=True)
        zf.writestr(str(model_metadata.model_id) + ".xml", buffer.getvalue())
        # Pack and rename proto-buffer file
        arcname = str(model_metadata.model_id) + ".model"
        if isinstance(model, str):
            zf.write(model, arcname=arcname)
        else:
            zf.writestr(arcname, model)
        # Pack license file
        if model_metadata.license_file is not None:
            zf.write(
                model_metadata.license_file,
                arcname=os.path.split(model_metadata.license_file)[1],
            )
        # Write empty file with model id
        zf.writestr("modelid=" + str(model_metadata.model_id), "")



# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Holds all relevant information for packaging and publishing to PyPI."""
import setuptools

# pylint: disable=line-too-long
with open("README.md", "r") as fh_read:
    long_description = fh_read.read()
setuptools.setup(
    name='czmodel',
    version='1.0.0',
    entry_points={'console_scripts': ['savedmodel2czmodel=czmodel.convert:main']},
    author="Sebastian Soyer",
    author_email="sebastian.soyer@zeiss.com",
    description="A conversion tool for TensorFlow ANNs to CZModel",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # Note: Exclude test folder in MANIFEST.in to also remove from source dist
    # See https://stackoverflow.com/questions/8556996/setuptools-troubles-excluding-packages-including-data-files
    # See https://docs.python.org/3.6/distutils/sourcedist.html
    packages=setuptools.find_packages(where='.', exclude=["test"]),
    license_files=['LICENSE.txt', 'NOTICE.txt'],
    # Classifiers help users find your project by categorizing it.
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    # Make required Python version compliant with official TF docs (https://www.tensorflow.org/install)
    # It follows: We build a pure Python wheel
    # Note that Python used to build the sources specifies the Python version of the dist (relevant during build only)
    # See https://packaging.python.org/guides/distributing-packages-using-setuptools/#pure-python-wheels for more info
    # We also restrict the code to >=3.6 to fully benefit from type annotations
    # See https://realpython.com/python-type-checking/ for more info
    python_requires='>=3.6,<4',
    # Exclude TF2.1 due to buggy behaviour
    install_requires=['tensorflow>=2,<2.3,!=2.1', 'onnx', 'tf2onnx'],
    # List additional URLs that are relevant to your project as a dict.
    # The key is what's used to render the link text on PyPI.
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    project_urls={
        'Demo': 'https://notebooks.azure.com/sebastian-soyer/projects/czmodel',
        'ZEN Intellesis': 'https://www.zeiss.com/microscopy/int/products/microscope-software/zen-intellesis-image-segmentation-by-deep-learning.html'
    },
)



"""Provides conversion functions to create a CZModel from exported TF models and corresponding meta data."""
from .convert import convert_from_json_spec, convert_from_model_spec
from .model_metadata import ModelSpec, ModelMetadata, SegmentationClass



# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides conversion utility functions."""
import logging
import os
import tempfile
from io import BytesIO
from typing import Sequence, Tuple, Union, Optional, Dict, TYPE_CHECKING
from zipfile import ZipFile

import onnx  # type: ignore  # Library does not provide type hints
import tensorflow as tf
from tf2onnx import tf_loader, utils, optimizer  # type: ignore  # Library does not provide type hints
from tf2onnx.tfonnx import process_tf_graph  # type: ignore  # Library does not provide type hints

from czmodel.util.model_packing import create_model_zip, zip_directory
from .preprocessing import add_preprocessing_layers

# Ignore flake8 imported but unused warnings (does not recognize non-runtime check). Checked by mypy
if TYPE_CHECKING:
    from tensorflow.keras import Model  # noqa: F401
    from tensorflow.keras.layers import Layer  # noqa: F401
    from czmodel.model_metadata import ModelMetadata  # noqa: F401 # pylint: disable=C0412


_ONNX_OPSET = 12


def convert_saved_model_to_onnx(
    saved_model_dir: str,
    output_file: str,
    target: Optional[Sequence] = None,
    custom_op_handlers: Optional[Dict] = None,
    extra_opset: Optional[Sequence] = None,
) -> None:
    """Exports a SavedModel on disk to an ONNX file.

    Arguments:
        saved_model_dir: The directory containing the model in SavedModel format.
        output_file: The path to the file to be created for the ONNX model.
        target: Workarounds applied to help certain platforms.
        custom_op_handlers: Handlers for custom operations.
        extra_opset: Extra opset, for example the opset used by custom ops.
    """
    graph_def, inputs, outputs = tf_loader.from_saved_model(saved_model_dir, None, None)

    with tf.Graph().as_default() as tf_graph:
        tf.import_graph_def(graph_def, name="")

    with tf_loader.tf_session(graph=tf_graph):
        g = process_tf_graph(
            tf_graph,
            continue_on_error=False,
            target=target,
            opset=_ONNX_OPSET,
            custom_op_handlers=custom_op_handlers,
            extra_opset=extra_opset,
            input_names=inputs,
            output_names=outputs,
        )

    onnx_graph = optimizer.optimize_graph(g)

    model_proto = onnx_graph.make_model(
        "Model {} in ONNX format".format(os.path.split(saved_model_dir)[1])
    )

    utils.save_protobuf(output_file, model_proto)

    # Verify model's structure and check for valid model schema
    onnx.checker.check_model(output_file)


def convert(
    model: "Model",
    model_metadata: "ModelMetadata",
    output_path: str,
    spatial_dims: Optional[Tuple[int, int]] = None,
    preprocessing: Union["Layer", Sequence["Layer"]] = None,
) -> None:
    """Converts a given Keras model to either an ONNX model or (upon failure) to a TensorFlow .saved_model.

    The exported model is optimized for inference.

    Args:
        model: Keras model to be converted. The model must have a separate InputLayer as input node.
        model_metadata: The metadata required to generate a CZModel.
        output_path: Destination path to the .czmodel file that will be generated.
        preprocessing: The layers to be prepended.
        spatial_dims: New spatial dimensions for the input node (see final usage for more details)
    """
    if preprocessing is not None or spatial_dims is not None:
        model = add_preprocessing_layers(
            model=model, layers=preprocessing, spatial_dims=spatial_dims
        )

    with tempfile.TemporaryDirectory() as tmpdir_saved_model_name, tempfile.TemporaryDirectory() as tmpdir_onnx_name:
        # Export Keras model in SavedModel format
        model.save(tmpdir_saved_model_name, include_optimizer=False, save_format="tf")
        try:
            # Convert to ONNX
            onnx_path = os.path.join(tmpdir_onnx_name, "model.onnx")
            convert_saved_model_to_onnx(tmpdir_saved_model_name, onnx_path)
            with open(onnx_path, "rb") as f:
                buffer = BytesIO(f.read())
        # pylint: disable=W0703  # Broad exception is justified here since this is a general fallback
        except Exception:
            # Log fallback to SavedModel
            logging.warning(
                "Model could not be converted to ONNX format. "
                "Falling back to SavedModel format. "
                "Using this format is discouraged."
            )
            # Zip saved model
            buffer = BytesIO()
            with ZipFile(buffer, mode="w") as zf:
                zip_directory(tmpdir_saved_model_name, zf)

        # Pack model into CZModel
        create_model_zip(
            model=buffer.getbuffer(),
            model_metadata=model_metadata,
            output_path=output_path,
        )



# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for input checking in argument parser."""
import argparse
import os


def dir_file(path: str) -> str:
    """Helper type for arg parser to check if passed argument is a valid file.

    Args:
        path: path to check

    Returns:
        path if valid

    Raises:
        ArgumentTypeError: path determined not to be a valid file
    """
    if os.path.isfile(path):
        return path

    raise argparse.ArgumentTypeError("Given argument is not a valid file")


def dir_path(path: str) -> str:
    """Helper type for arg parser to check if passed argument is a valid directory.

    Args:
        path: path to check

    Returns:
        path if valid

    Raises:
        ArgumentTypeError: path determined not to be a valid directory
    """
    if os.path.isdir(path):
        return path

    raise argparse.ArgumentTypeError("Given argument is not a valid directory")



# Copyright 2020 Carl Zeiss Microscopy GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides conversion utility functions."""
from io import BytesIO
from typing import TYPE_CHECKING

from czmodel.util.model_packing import create_model_zip

# Ignore flake8 imported but unused warnings (does not recognize non-runtime check). Checked by mypy
if TYPE_CHECKING:
    from czmodel.model_metadata import ModelMetadata  # noqa: F401


_ONNX_OPSET = 11


def convert(model: str, model_metadata: "ModelMetadata", output_path: str,) -> None:
    """Converts a given Keras model to a TensorFlow .pb model optimized for inference.

    Args:
        model: Keras model to be converted. The model must have a separate InputLayer as input node.
        model_metadata: The metadata required to generate a CZModel.
        output_path: Destination path to the .czmodel file that will be generated.
    """
    # Pack model into CZModel
    with open(model, "rb") as f:
        buffer = BytesIO(f.read())
        create_model_zip(
            model=buffer.getbuffer(),
            model_metadata=model_metadata,
            output_path=output_path,
        )
